# flake8: noqa

# pandas versioning
import pandas

# from packaging.version import Version
# pv = Version(pandas.__version__)

# if pv < '0.19.0':
#     raise ValueError("pandas_msgpack requires at least pandas 0.19.0")
# _is_pandas_legacy_version = pv.version[1] == 19 and len(pv.version) == 3

from .packers import to_msgpack, read_msgpack
from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("isf-pandas-msgpack")
except PackageNotFoundError:
    # package is not installed
    pass
# del get_versions, versions, pv, Version, pandas
del pandas



import pandas as pd
from packaging.version import Version
import sys

PANDAS_ST_200 = Version(pd.__version__).release < (2, 0, 0)
PANDAS_ST_210 = Version(pd.__version__).release < (2, 1, 0)
PANDAS_ST_120 = Version(pd.__version__).release < (1, 2, 0)
PANDAS_GE_210 = Version(pd.__version__).release >= (2, 1, 0)
PANDAS_GE_300 =  Version(pd.__version__).major >= 3
PY3 = (sys.version_info[0] >= 3)
STRING_TYPES = (str,)


if PANDAS_ST_120:
    from pandas.core.internals import _safe_reshape
    from pandas.io.common import get_filepath_or_buffer as _get_filepath_or_buffer
    def get_filepath_or_buffer(*args, **kwargs):
        fpb, encoding, compression, _ = _get_filepath_or_buffer(*args, **kwargs)
        return fpb, encoding, compression
else:
    from pandas.io.common import _get_filepath_or_buffer
    def get_filepath_or_buffer(*args, **kwargs):
        io_args = _get_filepath_or_buffer(*args, **kwargs)
        fpb, encoding, compression = io_args.filepath_or_buffer, io_args.encoding, io_args.compression
        return fpb, encoding, compression
    def _safe_reshape(arr, new_shape):
        """
        If possible, reshape `arr` to have shape `new_shape`,
        with a couple of exceptions (see gh-13012):

        1) If `arr` is a ExtensionArray or Index, `arr` will be
        returned as is.
        2) If `arr` is a Series, the `_values` attribute will
        be reshaped and returned.

        Parameters
        ----------
        arr : array-like, object to be reshaped
        new_shape : int or tuple of ints, the new shape
        """
        if isinstance(arr, ABCSeries):
            arr = arr._values
        if not is_extension_array_dtype(arr.dtype):
            # Note: this will include TimedeltaArray and tz-naive DatetimeArray
            # TODO(EA2D): special case will be unnecessary with 2D EAs
            arr = np.asarray(arr).reshape(new_shape)
        return arr

if PANDAS_ST_210:
    from pandas.core.arrays.sparse import SparseDtype
else:
    from pandas import SparseDtype

    
if PY3:
    def u(s):
        return s

    def u_safe(s):
        return s
else:
    def u(s):
        return unicode(s, "unicode_escape")

    def u_safe(s):
        try:
            return unicode(s, "unicode_escape")
        except:
            return s


from . import encode_decode, check_arbitrary, to_msgpack, read_msgpack
import pytest
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from packaging.version import Version


# class TestNDFrame(TestPackers):

def test_basic_frame(pandas_dataframe):
    for s, i in pandas_dataframe.items():
        i_rec = encode_decode(None, i)
        assert_frame_equal(i, i_rec)

@pytest.mark.skipif(
    Version(pd.__version__) < Version('1.3'), 
    reason="Prior to pandas 1.3, timezones were saved as a data attribute. This has since changed to be part of the dtype")
def test_datetime(pandas_datetime_dataframe):
    i_rec = encode_decode(None, pandas_datetime_dataframe)
    for k in pandas_datetime_dataframe.keys():
        assert_frame_equal(pandas_datetime_dataframe[k], i_rec[k])

def test_multi(pandas_dataframe):
    i_rec = encode_decode(None, pandas_dataframe)
    for k in pandas_dataframe.keys():
        assert_frame_equal(pandas_dataframe[k], i_rec[k])

    l = tuple([pandas_dataframe['float'], pandas_dataframe['float'].A,
                pandas_dataframe['float'].B, None])
    l_rec = encode_decode(None, l)
    check_arbitrary(l, l_rec)

    # this is an oddity in that packed lists will be returned as tuples
    l = [pandas_dataframe['float'], pandas_dataframe['float']
            .A, pandas_dataframe['float'].B, None]
    l_rec = encode_decode(None, l)
    assert isinstance(l_rec, tuple)
    check_arbitrary(l, l_rec)

def test_iterator(tmp_file_msg, pandas_dataframe):

    l = [pandas_dataframe['float'], pandas_dataframe['float']
            .A, pandas_dataframe['float'].B, None]

    to_msgpack(tmp_file_msg, *l)
    for i, packed in enumerate(read_msgpack(tmp_file_msg, iterator=True)):
        check_arbitrary(packed, l[i])

def tests_datetimeindex_freq_issue():

    # GH 5947
    # inferring freq on the datetimeindex
    df = pd.DataFrame([1, 2, 3], index=pd.date_range('1/1/2013', '1/3/2013'))
    result = encode_decode(None, df)
    assert_frame_equal(result, df)

    df = pd.DataFrame([1, 2], index=pd.date_range('1/1/2013', '1/2/2013'))
    result = encode_decode(None, df)
    assert_frame_equal(result, df)

def test_dataframe_duplicate_column_names():

    # GH 9618
    expected_1 = pd.DataFrame(columns=['a', 'a'])
    expected_2 = pd.DataFrame(columns=[1] * 100)
    expected_2.loc[0] = np.random.randn(100)
    expected_3 = pd.DataFrame(columns=[1, 1])
    expected_3.loc[0] = ['abc', np.nan]

    result_1 = encode_decode(None, expected_1)
    result_2 = encode_decode(None, expected_2)
    result_3 = encode_decode(None, expected_3)

    assert_frame_equal(result_1, expected_1)
    assert_frame_equal(result_2, expected_2)
    assert_frame_equal(result_3, expected_3)


from . import encode_decode, make_string_series
import pandas as pd
import numpy as np
from pandas.testing import (
    # assert_frame_equal,
    assert_series_equal,
)
import pytest

def _check_roundtrip(self, obj, comparator, **kwargs):

    # currently these are not implemetned
    i_rec = encode_decode(obj)
    # comparator(obj, i_rec, **kwargs)
    raise ValueError("{}\n{}".format(obj, pd.Series(i_rec)))
    with pytest.raises(NotImplementedError):
        encode_decode(obj)

@pytest.mark.skip(reason="Sparse Series are not implemented (yet): see packers.unconvert for reconsturcting sparse series.")
def test_sparse_series(self):

    s = make_string_series()
    s[3:5] = np.nan
    ss = s.astype("Sparse")
    self._check_roundtrip(ss, assert_series_equal, check_series_type=True)

    ss2 = s.astype('Sparse[int]')
    self._check_roundtrip(ss2, assert_series_equal, check_series_type=True)


# @pytest.mark.skip(reason="Sparse DataFrames are deprecated. Dataframes can however contain sparse Series.")
# def test_sparse_frame(self):

#     s = tm.makeDataFrame()
#     s.loc[3:5, 1:3] = np.nan
#     s.loc[8:10, -2] = np.nan
#     ss = s.astype("Sparse")

#     self._check_roundtrip(ss, tm.assert_frame_equal,
#                             check_frame_type=True)

#     ss2 = s.astype('Sparse[int]')
#     self._check_roundtrip(ss2, tm.assert_frame_equal,
#                             check_frame_type=True)

    # ss3 = s.to_sparse(fill_value=0)
    # self._check_roundtrip(ss3, tm.assert_frame_equal,
    #                       check_frame_type=True)


import pytest
import pandas as pd
import numpy as np
from . import to_msgpack, read_msgpack
from pandas.testing import assert_frame_equal

def test_string_io(tmp_file_msg):

    df = pd.DataFrame(np.random.randn(10, 2))

    s = to_msgpack(None, df)
    result = read_msgpack(s)
    assert_frame_equal(result, df)

    # s = to_msgpack(None, df)
    # result = read_msgpack(s)
    # tm.assert_frame_equal(result, df)

    to_msgpack(str(tmp_file_msg), df)
    with open(tmp_file_msg, 'rb') as fh:
        result = read_msgpack(s)
    assert_frame_equal(result, df)

    # s = to_msgpack(self.path, df)
    # result = read_msgpack(s)
    # tm.assert_frame_equal(result, df)

    s = to_msgpack(None, df)
    with open(tmp_file_msg, 'wb') as fh:
        fh.write(s)
    result = read_msgpack(tmp_file_msg)
    assert_frame_equal(result, df)

def test_iterator_with_string_io():

    dfs = [pd.DataFrame(np.random.randn(10, 2)) for i in range(5)]
    s = to_msgpack(None, *dfs)
    for i, result in enumerate(read_msgpack(s, iterator=True)):
        assert_frame_equal(result, dfs[i])

def test_invalid_arg():
    # GH10369
    class A(object):
        def __init__(self):
            self.read = 0

    with pytest.raises(ValueError):
        read_msgpack(path_or_buf=None)
    with pytest.raises(ValueError):
        read_msgpack(path_or_buf={})
    with pytest.raises(ValueError):
        read_msgpack(path_or_buf=A())


"""See https://github.com/pandas-dev/pandas/pull/9783
"""
from . import encode_decode
import pandas as pd
from pandas.testing import assert_frame_equal
from pandas.errors import PerformanceWarning
import pytest
import numpy as np
from unittest.mock import patch
from numpy.testing import assert_array_equal as assert_numpy_array_equal

try:
    from sqlalchemy import create_engine
    _create_sql_engine = create_engine
except ImportError:
    _SQLALCHEMY_INSTALLED = False
else:
    _SQLALCHEMY_INSTALLED = True
try:
    import blosc  # NOQA
except ImportError:
    _BLOSC_INSTALLED = False
else:
    _BLOSC_INSTALLED = True

try:
    import zlib  # NOQA
except ImportError:
    _ZLIB_INSTALLED = False
else:
    _ZLIB_INSTALLED = True


def test_plain(pandas_dataframe):
    i_rec = encode_decode(None, pandas_dataframe)
    for k in pandas_dataframe.keys():
        assert_frame_equal(pandas_dataframe[k], i_rec[k])

def _test_compression(pandas_dataframe, compress):
    i_rec = encode_decode(None, pandas_dataframe, compress=compress)
    for k in pandas_dataframe.keys():
        value = i_rec[k]
        expected = pandas_dataframe[k]
        assert_frame_equal(value, expected)
        # make sure that we can write to the new frames
        for block in value._mgr.blocks:
            if isinstance(block.values, np.ndarray):
                assert block.values.flags.writeable
            else:
                # e.g. DatetimeArray
                # Access underlying _data array
                if hasattr(block.values, '_data'):
                    assert block.values._data.flags.writeable
                else:
                    # DatetimeArrays are still experimental with changing API - just pass... - Bjorge 2025-03-24
                    pass

@pytest.mark.skipif(not _ZLIB_INSTALLED, reason="zlib not installed")
def test_compression_zlib(pandas_dataframe):
    _test_compression(pandas_dataframe, 'zlib')

@pytest.mark.skipif(not _BLOSC_INSTALLED, reason="blosc not installed")
def test_compression_blosc(pandas_dataframe):
    _test_compression(pandas_dataframe, 'blosc')

def _test_compression_warns_when_decompress_caches(pandas_dataframe, compress):
    not_garbage = []
    control = []  # copied data

    compress_module = globals()[compress]
    real_decompress = compress_module.decompress

    def decompress(ob):
        """mock decompress function that delegates to the real
        decompress but caches the result and a copy of the result.
        """
        res = real_decompress(ob)
        not_garbage.append(res)  # hold a reference to this bytes object
        control.append(bytearray(res))  # copy the data here to check later
        return res

    # types mapped to values to add in place.
    rhs = {
        np.dtype('float64'): 1.0,
        np.dtype('int32'): 1,
        np.dtype('object'): 'a',
        np.dtype('datetime64[ns]'): np.timedelta64(1, 'ns'),
        np.dtype('timedelta64[ns]'): np.timedelta64(1, 'ns'),
    }

    with patch(f"{compress}.decompress", decompress), pytest.raises(PerformanceWarning) as excinfo:

        i_rec = encode_decode(
            None,
            pandas_dataframe, 
            compress=compress
            )
        for k in pandas_dataframe.keys():
            value = i_rec[k]
            expected = pandas_dataframe[k]
            assert_frame_equal(value, expected)
            # make sure that we can write to the new frames even though
            # we needed to copy the data
            for block in value._mgr.blocks:
                if isinstance(block.values, np.ndarray):
                    assert block.values.flags.writeable
                    # mutate the data in some way
                    block.values[0] += rhs[block.dtype]
                else:
                    # e.g. DatetimeArray
                    # Access underlying _data array
                    assert block.values._data.flags.writeable
                    block.values._data[0] += rhs[block.dtype]

    assert excinfo.match('copying data after decompressing; this may mean that decompress is caching its result')
    for buf, control_buf in zip(not_garbage, control):
        # make sure none of our mutations above affected the
        # original buffers
        assert buf == control_buf

@pytest.mark.skipif(not _ZLIB_INSTALLED, reason="zlib not installed")
def test_compression_warns_when_decompress_caches_zlib(pandas_dataframe):
    _test_compression_warns_when_decompress_caches(pandas_dataframe, 'zlib')

@pytest.mark.skipif(not _BLOSC_INSTALLED, reason="blosc not installed")
def test_compression_warns_when_decompress_caches_blosc(pandas_dataframe):
    _test_compression_warns_when_decompress_caches(pandas_dataframe, 'blosc')

def _test_small_strings_no_warn(compress):
    empty = np.array([], dtype='uint8')
    empty_unpacked = encode_decode(None, empty, compress=compress)

    assert_numpy_array_equal(empty_unpacked, empty)
    # Ensure that the unpacked array is writable
    assert empty_unpacked.flags.writeable

    char = np.array([ord(b'a')], dtype='uint8')
    char_unpacked = encode_decode(None, char, compress=compress)

    assert_numpy_array_equal(char_unpacked, char)
    assert char_unpacked.flags.writeable
    # if this test fails I am sorry because the interpreter is now in a
    # bad state where b'a' points to 98 == ord(b'b').
    char_unpacked[0] = ord(b'b')

    # we compare the ord of bytes b'a' with unicode u'a' because the should
    # always be the same (unless we were able to mutate the shared
    # character singleton in which case ord(b'a') == ord(b'b').
    assert ord(b'a') == ord(u'a')
    assert_numpy_array_equal(
        char_unpacked,
        np.array([ord(b'b')], dtype='uint8'),
    )

@pytest.mark.skipif(not _ZLIB_INSTALLED,reason="zlib not installed")
def test_small_strings_no_warn_zlib():
    _test_small_strings_no_warn('zlib')

@pytest.mark.skipif(not _BLOSC_INSTALLED, reason="blosc not installed")
def test_small_strings_no_warn_blosc():
    _test_small_strings_no_warn('blosc')

@pytest.mark.skipif(not _BLOSC_INSTALLED, reason="blosc not installed")
def test_readonly_axis_blosc():
    # GH11880
    df1 = pd.DataFrame({'A': list('abcd')})
    df2 = pd.DataFrame(df1, index=[1., 2., 3., 4.])
    assert 1 in encode_decode(None, df1['A'], compress='blosc')
    assert 1. in encode_decode(None, df2['A'], compress='blosc')

def test_readonly_axis_zlib():
    # GH11880
    df1 = pd.DataFrame({'A': list('abcd')})
    df2 = pd.DataFrame(df1, index=[1., 2., 3., 4.])
    assert 1 in encode_decode(None, df1['A'], compress='zlib')
    assert 1. in encode_decode(None, df2['A'], compress='zlib')

@pytest.mark.skipif(not _SQLALCHEMY_INSTALLED, reason="sqlalchemy not installed")
@pytest.mark.skipif(not _BLOSC_INSTALLED, reason="blosc not installed")
def test_readonly_axis_blosc_to_sql():
    # GH11880
    expected = pd.DataFrame({'A': list('abcd')})
    df = encode_decode(None, expected, compress='blosc')
    eng = _create_sql_engine("sqlite:///:memory:")
    df.to_sql('test', eng, if_exists='append')
    result = pd.read_sql_table('test', eng, index_col='index')
    result.index.names = [None]
    assert_frame_equal(expected, result)


@pytest.mark.skipif(not _SQLALCHEMY_INSTALLED, reason="sqlalchemy not installed")
@pytest.mark.skipif(not _ZLIB_INSTALLED, reason="zlib not installed")
def test_readonly_axis_zlib_to_sql():
    # GH11880
    expected = pd.DataFrame({'A': list('abcd')})
    df = encode_decode(None, expected, compress='zlib')
    eng = _create_sql_engine("sqlite:///:memory:")
    df.to_sql('test', eng, if_exists='append')
    result = pd.read_sql_table('test', eng, index_col='index')
    result.index.names = [None]
    assert_frame_equal(expected, result)


"""
Msgpack serializer support for reading and writing pandas data structures
to disk
portions of msgpack_numpy package, by Lev Givon were incorporated
into this module (and tests_packers.py)
License
=======
Copyright (c) 2013, Lev Givon.
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.
* Neither the name of Lev Givon nor the names of any
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os, io
import numpy as np
from datetime import datetime, date, timedelta
from dateutil.parser import parse
from textwrap import dedent

try:
    from pandas.core.dtypes.common import (
        # is_categorical_dtype, 
        is_object_dtype, 
        needs_i8_conversion, 
        pandas_dtype, 
        is_extension_array_dtype,
        PeriodDtype
        )
except ImportError:
    from pandas.types.common import (
        # is_categorical_dtype, 
        is_object_dtype, 
        needs_i8_conversion, 
        pandas_dtype, 
        is_extension_array_dtype,
        PeriodDtype
        )

from pandas import (
    Timestamp, 
    Period, 
    Series, 
    DataFrame,  # noqa
    Index, 
    MultiIndex,
    RangeIndex, 
    PeriodIndex, 
    DatetimeIndex, 
    NaT,
    Categorical, 
    CategoricalIndex, 
    CategoricalDtype
)
from pandas.core.arrays.sparse.array import BlockIndex, IntIndex
from pandas.arrays import PeriodArray
from pandas.core.generic import NDFrame
from pandas.core.dtypes.generic import ABCSeries
from pandas.core.dtypes.common import is_extension_array_dtype
from pandas.errors import PerformanceWarning
from .pandas_compat import (
    get_filepath_or_buffer, 
    SparseDtype, 
    PANDAS_GE_300, 
    PANDAS_GE_210,
    u,
    u_safe,
    STRING_TYPES
    )

from isf_pandas_msgpack.msgpack import (
    Unpacker as _Unpacker,
    Packer as _Packer,
    ExtType
)
from isf_pandas_msgpack._move import (
    BadMove as _BadMove,
    move_into_mutable_buffer as _move_into_mutable_buffer,
)
NaTType = type(NaT)

# check which compression libs we have installed
try:
    import zlib
    def _check_zlib():
        pass
except ImportError:
    def _check_zlib():
        raise ImportError('zlib is not installed')

try:
    import blosc
    def _check_blosc():
        pass
except ImportError:
    def _check_blosc():
        raise ImportError('blosc is not installed')

# until we can pass this into our conversion functions,
# this is pretty hacky
compressor = None

dtype_dict = {
    21: np.dtype('M8[ns]'),
    u('datetime64[ns]'): np.dtype('M8[ns]'),
    u('datetime64[us]'): np.dtype('M8[us]'),
    22: np.dtype('m8[ns]'),
    u('timedelta64[ns]'): np.dtype('m8[ns]'),
    u('timedelta64[us]'): np.dtype('m8[us]'),

    # this is platform int, which we need to remap to np.int64
    # for compat on windows platforms
    7: np.dtype('int64'),
    'category': 'category'
    }



def _safe_reshape(arr, new_shape):
    """
    If possible, reshape `arr` to have shape `new_shape`,
    with a couple of exceptions (see gh-13012):

    1) If `arr` is a ExtensionArray or Index, `arr` will be
       returned as is.
    2) If `arr` is a Series, the `_values` attribute will
       be reshaped and returned.

    Parameters
    ----------
    arr : array-like, object to be reshaped
    new_shape : int or tuple of ints, the new shape
    """
    if isinstance(arr, ABCSeries):
        arr = arr._values
    if not is_extension_array_dtype(arr.dtype):
        # Note: this will include TimedeltaArray and tz-naive DatetimeArray
        # TODO(EA2D): special case will be unnecessary with 2D EAs
        arr = np.asarray(arr).reshape(new_shape)
    return arr


def to_msgpack(path_or_buf, *args, **kwargs):
    """
    msgpack (serialize) object to input file path
    Parameters
    ----------
    path_or_buf : string File path, buffer-like, or None
                  if None, return generated string
    args : an object or objects to serialize
    encoding: encoding for unicode objects
    append : boolean whether to append to an existing msgpack
             (default is False)
    compress : type of compressor (zlib or blosc), default to None (no
               compression)
    """
    global compressor
    compressor = kwargs.pop('compress', None)
    if compressor:
        compressor = u(compressor)
    append = kwargs.pop('append', None)
    if append:
        mode = 'a+b'
    else:
        mode = 'wb'

    def writer(fh):
        for a in args:
            fh.write(pack(a, **kwargs))

    if isinstance(path_or_buf, STRING_TYPES):
        with open(path_or_buf, mode) as fh:
            writer(fh)
    elif path_or_buf is None:
        buf = io.BytesIO()
        writer(buf)
        return buf.getvalue()
    else:
        writer(path_or_buf)


def read_msgpack(path_or_buf, encoding='utf-8', iterator=False, **kwargs):
    """
    Load msgpack pandas object from the specified
    file path
    Parameters
    ----------
    path_or_buf : string File path, BytesIO like or string
    encoding: Encoding for decoding msgpack str type
    iterator : boolean, if True, return an iterator to the unpacker
               (default is False)
    Returns
    -------
    obj : type of object stored in file
    """
    path_or_buf, _, _ = get_filepath_or_buffer(path_or_buf)
    if iterator:
        return Iterator(path_or_buf)

    def read(fh):
        l = list(unpack(fh, encoding=encoding, **kwargs))
        if len(l) == 1:
            return l[0]
        return l

    # see if we have an actual file
    if isinstance(path_or_buf, STRING_TYPES):

        try:
            exists = os.path.exists(path_or_buf)
        except (TypeError, ValueError):
            exists = False

        if exists:
            with open(path_or_buf, 'rb') as fh:
                return read(fh)

    # treat as a binary-like
    if isinstance(path_or_buf, bytes):
        fh = None
        try:
            fh = io.BytesIO(path_or_buf)
            return read(fh)
        finally:
            if fh is not None:
                fh.close()

    # a buffer like
    if hasattr(path_or_buf, 'read') and callable(path_or_buf.read):
        return read(path_or_buf)

    raise ValueError('path_or_buf needs to be a string file path or file-like')


def dtype_for(t):
    """ return my dtype mapping, whether number or name """
    if t in dtype_dict:
        return dtype_dict[t]
    return np.sctypeDict.get(t, t)


c2f_dict = {'complex': np.float64,
            'complex128': np.float64,
            'complex64': np.float32}

# numpy 1.6.1 compat
if hasattr(np, 'float128'):
    c2f_dict['complex256'] = np.float128


def c2f(r, i, ctype_name):
    """
    Convert strings to complex number instance with specified numpy type.
    """

    ftype = c2f_dict[ctype_name]
    return np.sctypeDict[ctype_name](ftype(r) + 1j * ftype(i))


def convert(values):
    """ convert the numpy values to a list """

    dtype = values.dtype

    if isinstance(dtype, CategoricalDtype):
        return values

    # if is_categorical_dtype(values):
    #     return values

    elif is_object_dtype(dtype):
        return values.ravel().tolist()

    if needs_i8_conversion(dtype):
        values = values.view('i8')
    v = values.ravel()

    if compressor == 'zlib':
        _check_zlib()

        # return string arrays like they are
        if dtype == np.object_:
            return v.tolist()

        # convert to a bytes array
        v = v.tobytes()
        return ExtType(0, zlib.compress(v))

    elif compressor == 'blosc':
        _check_blosc()

        # return string arrays like they are
        if dtype == np.object_:
            return v.tolist()

        # convert to a bytes array
        v = v.tobytes()
        return ExtType(0, blosc.compress(v, typesize=dtype.itemsize))

    # ndarray (on original dtype)
    return ExtType(0, v.tobytes())


def unconvert(values, dtype, compress=None):

    as_is_ext = isinstance(values, ExtType) and values.code == 0

    if as_is_ext:
        values = values.data

    if isinstance(values, Categorical):
        return values
    
    # if is_categorical_dtype(dtype):
    #     return values

    elif is_object_dtype(dtype):
        return np.array(values, dtype=object)

    original_dtype = pandas_dtype(dtype)
    if isinstance(original_dtype, SparseDtype):
        dtype = original_dtype.subtype
    elif isinstance(original_dtype, PeriodDtype):
        dtype = np.int64
    else:
        dtype = original_dtype.base

    if not as_is_ext:
        values = values.encode('latin1')

    if compress:
        if compress == u'zlib':
            _check_zlib()
            decompress = zlib.decompress
        elif compress == u'blosc':
            _check_blosc()
            decompress = blosc.decompress
        else:
            raise ValueError("compress must be one of 'zlib' or 'blosc'")

        try:
            array = np.frombuffer(
                _move_into_mutable_buffer(decompress(values)),
                dtype=dtype,
            )
        except _BadMove as e:
            # Pull the decompressed data off of the `_BadMove` exception.
            # We don't just store this in the locals because we want to
            # minimize the risk of giving users access to a `bytes` object
            # whose data is also given to a mutable buffer.
            values = e.args[0]
            if len(values) > 1:
                # The empty string and single characters are memoized in many
                # string creating functions in the capi. This case should not
                # warn even though we need to make a copy because we are only
                # copying at most 1 byte.
                raise PerformanceWarning(
                    'copying data after decompressing; this may mean that'
                    ' decompress is caching its result',
                    PerformanceWarning,
                )
            # fall through to copying `np.fromstring`
            array = np.frombuffer(values, dtype=dtype)
    else:
        array = np.frombuffer(values, dtype=dtype)

    # Set array to be writeable
    if not array.flags.writeable:
        # Move the data into a new array
        # This is identical to how the deprecated np.fromstring worked
        # in the old version of packers.
        array = array.copy()
        array.setflags(write=True)
    
    # Convert to PeriodArray if dtype is PeriodDtype
    if isinstance(original_dtype, PeriodDtype):
        # array = PeriodArray(array, freq=original_dtype.freq)
        array = PeriodArray(array, dtype=original_dtype)
        
    return array


def encode(obj):
    """
    Data encoder
    """
    tobj = type(obj)
    if isinstance(obj, Index):
        if isinstance(obj, RangeIndex):
            return {u'typ': u'range_index',
                    u'klass': u(obj.__class__.__name__),
                    u'name': getattr(obj, 'name', None),
                    u'start': getattr(obj, 'start', None),
                    u'stop': getattr(obj, 'stop', None),
                    u'step': getattr(obj, 'step', None)}
        elif isinstance(obj, PeriodIndex):
            return {u'typ': u'period_index',
                    u'klass': u(obj.__class__.__name__),
                    u'name': getattr(obj, 'name', None),
                    u'freq': u_safe(getattr(obj, 'freqstr', None)),
                    u'dtype': u(obj.dtype.name),
                    u'data': convert(obj.asi8),
                    u'compress': compressor}
        elif isinstance(obj, DatetimeIndex):
            tz = getattr(obj, 'tz', None)

            # store tz info and data as UTC
            if tz is not None:
                tz = u(tz.zone)
                obj = obj.tz_convert('UTC')
            return {u'typ': u'datetime_index',
                    u'klass': u(obj.__class__.__name__),
                    u'name': getattr(obj, 'name', None),
                    u'dtype': u(obj.dtype.name),
                    u'data': convert(obj.asi8),
                    u'freq': u_safe(getattr(obj, 'freqstr', None)),
                    u'tz': tz,
                    u'compress': compressor}
        elif isinstance(obj, MultiIndex):
            return {u'typ': u'multi_index',
                    u'klass': u(obj.__class__.__name__),
                    u'names': getattr(obj, 'names', None),
                    u'dtype': u(obj.dtype.name),
                    u'data': convert(obj.values),
                    u'compress': compressor}
        else:
            return {u'typ': u'index',
                    u'klass': u(obj.__class__.__name__),
                    u'name': getattr(obj, 'name', None),
                    u'dtype': u(obj.dtype.name),
                    u'data': convert(obj.values),
                    u'compress': compressor}

    elif isinstance(obj, Categorical):
        return {u'typ': u'category',
                u'klass': u(obj.__class__.__name__),
                u'name': getattr(obj, 'name', None),
                u'codes': obj.codes,
                u'categories': obj.categories,
                u'ordered': obj.ordered,
                u'compress': compressor}

    elif isinstance(obj, Series):
        if isinstance(obj.dtype, SparseDtype):
            d = {'typ': 'sparse_series',
                'klass': obj.__class__.__name__,
                'dtype': obj.dtype.name,
                'index': obj.index,
                'sp_index': obj.values.sp_index,
                'sp_values': convert(obj.values.sp_values),
                'compress': compressor}
            for f in ['name', 'fill_value', 'kind']:
               d[f] = getattr(obj, f, None)
            return d
        else:
            return {u'typ': u'series',
                    u'klass': u(obj.__class__.__name__),
                    u'name': getattr(obj, 'name', None),
                    u'index': obj.index,
                    u'dtype': u(obj.dtype.name),
                    u'data': convert(obj.values),
                    u'compress': compressor}
    elif issubclass(tobj, NDFrame):
        # if isinstance(obj.dtype, SparseDtype):
        #     d = {'typ': 'sparse_dataframe',
        #         'klass': obj.__class__.__name__,
        #         'columns': obj.columns}
        #     for f in ['default_fill_value', 'default_kind']:
        #        d[f] = getattr(obj, f, None)
        #     d['data'] = dict([(name, ss)
        #                     for name, ss in compat.iteritems(obj)])
        #     return d
        # else:

        # # The private _data attribute will be deprecated in the future. here, we use the BlockManager _mgr instead - Bjorge 2025-03-25
        data = obj._mgr
        if not data.is_consolidated():
            data = data.consolidate()

        # the block manager
        return {u'typ': u'block_manager',
                u'klass': u(obj.__class__.__name__),
                u'axes': data.axes,
                u'blocks': [
                    {
                        u'locs': b.mgr_locs.as_array,
                        u'values': convert(b.values),
                        u'shape': b.values.shape,
                        u'dtype': u(b.dtype.name),
                        u'klass': u(b.__class__.__name__),
                        u'compress': compressor} 
                    for b in data.blocks]
                }

    elif isinstance(obj, (datetime, date, np.datetime64, timedelta,
                          np.timedelta64, NaTType)):
        if isinstance(obj, Timestamp):
            tz = obj.tzinfo
            if tz is not None:
                tz = u(tz.zone)
            # Handle freq attribute for pandas versions < 2.0
            enc_obj = {u'typ': u'timestamp',
                    u'value': obj.value,
                    u'tz': tz}
            freq = getattr(obj, 'freq', None)
            if freq is not None:
                freq = u(freq.freqstr)
                enc_obj[u'freq'] = freq
            return enc_obj
        if isinstance(obj, NaTType):
            return {u'typ': u'nat'}
        elif isinstance(obj, np.timedelta64):
            return {u'typ': u'timedelta64',
                    u'data': obj.view('i8')}
        elif isinstance(obj, timedelta):
            return {u'typ': u'timedelta',
                    u'data': (obj.days, obj.seconds, obj.microseconds)}
        elif isinstance(obj, np.datetime64):
            return {u'typ': u'datetime64',
                    u'data': u(str(obj))}
        elif isinstance(obj, datetime):
            return {u'typ': u'datetime',
                    u'data': u(obj.isoformat())}
        elif isinstance(obj, date):
            return {u'typ': u'date',
                    u'data': u(obj.isoformat())}
        raise Exception("cannot encode this datetimelike object: %s" % obj)
    elif isinstance(obj, Period):
        return {u'typ': u'period',
                u'ordinal': obj.ordinal,
                u'freq': u(obj.freq)}
    elif isinstance(obj, BlockIndex):
        return {u'typ': u'block_index',
                u'klass': u(obj.__class__.__name__),
                u'blocs': obj.blocs,
                u'blengths': obj.blengths,
                u'length': obj.length}
    elif isinstance(obj, IntIndex):
        return {u'typ': u'int_index',
                u'klass': u(obj.__class__.__name__),
                u'indices': obj.indices,
                u'length': obj.length}
    elif isinstance(obj, np.ndarray):
        return {u'typ': u'ndarray',
                u'shape': obj.shape,
                u'ndim': obj.ndim,
                u'dtype': u(obj.dtype.name),
                u'data': convert(obj),
                u'compress': compressor}
    elif isinstance(obj, np.number):
        # don't use __repr__, breaks in numpy > 2.0 as it no longer represents just the number - bjorge 2.25-03-24
        if np.iscomplexobj(obj):
            return {u'typ': u'np_scalar',
                    u'sub_typ': u'np_complex',
                    u'dtype': u(obj.dtype.name),
                    u'real': u(obj.real.item()),
                    u'imag': u(obj.imag.item())}
        else:
            return {u'typ': u'np_scalar',
                    u'dtype': u(obj.dtype.name),
                    u'data': u(obj.item())} 
    elif isinstance(obj, complex):
        return {u'typ': u'np_complex',
                u'real': u(obj.real.__repr__()),
                u'imag': u(obj.imag.__repr__())}

    return obj


def _create_block(b, axes):
    from pandas.core.internals import make_block
    import pandas.core.internals as internals
    
    # Dynamically resolve the block class
    # Old pandas versions had granular block types (e.g. "FloatBlock", "IntBlock" ...)
    # Newer pandas version infer the dtype from ... well, the dtype. - Bjorge 2025-03-26
    block_class_name = b[u'klass']
    block_class = getattr(internals.blocks, block_class_name, None)
    # For newer pandas versions, fallback to a generic block if the specific class doesn't exist
    if block_class is None:
        from pandas.core.internals.blocks import new_block
        block_class = new_block
        
    values = _safe_reshape(unconvert(
        b[u'values'], dtype_for(b[u'dtype']),
        b[u'compress']), b[u'shape'])

    # locs handles duplicate column names, and should be used instead
    # of items; see GH 9618
    if u'locs' in b:
        placement = b[u'locs']
    else:
        placement = axes[0].get_indexer(b[u'items'])

    return make_block(
        values=values,
        klass=block_class,
        placement=placement,
        dtype=b[u'dtype'])

    
def _construct_df_from_blocks(obj):
    axes = obj[u'axes']
    # if PANDAS_GE_300:
    #     from pandas.api.internals import create_dataframe_from_blocks
    #     create_dataframe_from_blocks(
    #         obj[u'blocks'], axes[1], axes[0]
    #     )
    # else:
    from pandas.core.internals import BlockManager
    blocks = [_create_block(b=b, axes=axes) for b in obj[u'blocks']]
    mgr = BlockManager(blocks, list(axes))
    if PANDAS_GE_210:
        return DataFrame._from_mgr(mgr, axes=axes)
    else:
        return DataFrame(BlockManager(blocks, list(axes))) 
        

def decode(obj):
    """
    Decoder for deserializing numpy data types.
    """

    typ = obj.get(u'typ')
    if typ is None:
        return obj
    elif typ == u'timestamp':
        if "freq" in obj or "offset" in obj:
            freq = obj[u'freq'] if 'freq' in obj else obj[u'offset']
            return Timestamp(obj[u'value'], tz=obj[u'tz'], freq=freq)
        else:
            return Timestamp(obj[u'value'], tz=obj[u'tz'])
    elif typ == u'nat':
        return NaT
    elif typ == u'period':
        return Period(ordinal=obj[u'ordinal'], freq=obj[u'freq'])
    elif typ == u'index':
        dtype = dtype_for(obj[u'dtype'])
        data = unconvert(obj[u'data'], dtype,
                         obj.get(u'compress'))
        # return globals()[obj[u'klass']](data, dtype=dtype, name=obj[u'name'])
        return Index(data, dtype=dtype, name=obj[u'name'])
    elif typ == u'range_index':
        return globals()[obj[u'klass']](
            obj[u'start'],
            obj[u'stop'],
            obj[u'step'],
            name=obj[u'name'])
    elif typ == u'multi_index':
        dtype = dtype_for(obj[u'dtype'])
        data = unconvert(obj[u'data'], dtype,
                         obj.get(u'compress'))
        data = [tuple(x) for x in data]
        return globals()[obj[u'klass']].from_tuples(data, names=obj[u'names'])
    elif typ == u'period_index':
        data = unconvert(obj[u'data'], obj[u'dtype'], obj.get(u'compress'))
        d = dict(name=obj[u'name'], freq=obj[u'freq'])
        # raise ValueError(obj)
        # if _is_pandas_legacy_version:
        #     # legacy
        #     return globals()[obj[u'klass']](data, **d)
        # else:
        #     return globals()[obj[u'klass']]._from_ordinals(data, **d)
        return globals()[obj[u'klass']](data, **d)
    elif typ == u'datetime_index':
        data = unconvert(obj[u'data'], np.int64, obj.get(u'compress'))
        d = dict(
            name=obj[u'name'], 
            freq=obj[u'freq'], 
            # verify_integrity=False
            )
        result = globals()[obj[u'klass']](data, **d)
        tz = obj[u'tz']

        # reverse tz conversion
        if tz is not None:
            result = result.tz_localize('UTC').tz_convert(tz)
        return result

    elif typ == u'category':
        from_codes = globals()[obj[u'klass']].from_codes
        return from_codes(codes=obj[u'codes'],
                          categories=obj[u'categories'],
                          ordered=obj[u'ordered'])

    elif typ == u'series':
        dtype = dtype_for(obj[u'dtype'])
        pd_dtype = pandas_dtype(dtype)

        index = obj[u'index']
        result = globals()[obj[u'klass']](unconvert(obj[u'data'], dtype,
                                                    obj[u'compress']),
                                          index=index,
                                          dtype=pd_dtype,
                                          name=obj[u'name'])
        return result

    elif typ == u'block_manager':
        return _construct_df_from_blocks(obj)
    elif typ == u'datetime':
        return parse(obj[u'data'])
    elif typ == u'datetime64':
        return np.datetime64(parse(obj[u'data']))
    elif typ == u'date':
        return parse(obj[u'data']).date()
    elif typ == u'timedelta':
        return timedelta(*obj[u'data'])
    elif typ == u'timedelta64':
        return np.timedelta64(int(obj[u'data']))
    # elif typ == 'sparse_series':
    #     dtype = dtype_for(obj['dtype'])
    #     sp_values = unconvert(obj['sp_values'], dtype, obj['compress'])
    #     index = obj['index']
    #     s = globals()[obj['klass']](sp_values, index=index, name=obj['name'])
    #     s.fill_value = obj['fill_value']
    #     s.kind = obj['kind']
    #     return s
    # elif typ == 'sparse_dataframe':
    #    return globals()[obj['klass']](
    #        obj['data'], columns=obj['columns'],
    #        default_fill_value=obj['default_fill_value'],
    #        default_kind=obj['default_kind']
    #    )
    # elif typ == 'sparse_panel':
    #    return globals()[obj['klass']](
    #        obj['data'], items=obj['items'],
    #        default_fill_value=obj['default_fill_value'],
    #        default_kind=obj['default_kind'])
    elif typ == u'block_index':
        return globals()[obj[u'klass']](obj[u'length'], obj[u'blocs'],
                                        obj[u'blengths'])
    elif typ == u'int_index':
        return globals()[obj[u'klass']](obj[u'length'], obj[u'indices'])
    elif typ == u'ndarray':
        return unconvert(obj[u'data'], np.sctypeDict[obj[u'dtype']],
                         obj.get(u'compress')).reshape(obj[u'shape'])
    elif typ == u'np_scalar':
        if obj.get(u'sub_typ') == u'np_complex':
            return c2f(obj[u'real'], obj[u'imag'], obj[u'dtype'])
        else:
            dtype = dtype_for(obj[u'dtype'])
            try:
                return dtype(obj[u'data'])
            except:
               return dtype.type(obj[u'data'])
    elif typ == u'np_complex':
        return complex(obj[u'real'] + u'+' + obj[u'imag'] + u'j')
    elif isinstance(obj, (dict, list, set)):
        return obj
    else:
        return obj


def pack(o, default=encode,
         encoding='utf-8', unicode_errors='strict', use_single_float=False,
         autoreset=1, use_bin_type=1):
    """
    Pack an object and return the packed bytes.
    """

    return Packer(default=default, encoding=encoding,
                  unicode_errors=unicode_errors,
                  use_single_float=use_single_float,
                  autoreset=autoreset,
                  use_bin_type=use_bin_type).pack(o)


def unpack(packed, object_hook=decode,
           list_hook=None, use_list=False, encoding='utf-8',
           unicode_errors='strict', object_pairs_hook=None,
           max_buffer_size=0, ext_hook=ExtType):
    """
    Unpack a packed object, return an iterator
    Note: packed lists will be returned as tuples
    """

    return Unpacker(packed, object_hook=object_hook,
                    list_hook=list_hook,
                    use_list=use_list, encoding=encoding,
                    unicode_errors=unicode_errors,
                    object_pairs_hook=object_pairs_hook,
                    max_buffer_size=max_buffer_size,
                    ext_hook=ext_hook)


class Packer(_Packer):

    def __init__(self, default=encode,
                 encoding='utf-8',
                 unicode_errors='strict',
                 use_single_float=False,
                 autoreset=1,
                 use_bin_type=1):
        super(Packer, self).__init__(default=default,
                                     encoding=encoding,
                                     unicode_errors=unicode_errors,
                                     use_single_float=use_single_float,
                                     autoreset=autoreset,
                                     use_bin_type=use_bin_type)


class Unpacker(_Unpacker):

    def __init__(self, file_like=None, read_size=0, use_list=False,
                 object_hook=decode,
                 object_pairs_hook=None, list_hook=None, encoding='utf-8',
                 unicode_errors='strict', max_buffer_size=0, ext_hook=ExtType):
        super(Unpacker, self).__init__(file_like=file_like,
                                       read_size=read_size,
                                       use_list=use_list,
                                       object_hook=object_hook,
                                       object_pairs_hook=object_pairs_hook,
                                       list_hook=list_hook,
                                       encoding=encoding,
                                       unicode_errors=unicode_errors,
                                       max_buffer_size=max_buffer_size,
                                       ext_hook=ext_hook)


class Iterator(object):

    """ manage the unpacking iteration,
        close the file on completion """

    def __init__(self, path, **kwargs):
        self.path = path
        self.kwargs = kwargs

    def __iter__(self):

        needs_closing = True
        try:

            # see if we have an actual file
            if isinstance(self.path, STRING_TYPES):

                try:
                    path_exists = os.path.exists(self.path)
                except TypeError:
                    path_exists = False

                if path_exists:
                    fh = open(self.path, 'rb')
                else:
                    fh = io.BytesIO(self.path)

            else:

                if not hasattr(self.path, 'read'):
                    fh = io.BytesIO(self.path)

                else:

                    # a file-like
                    needs_closing = False
                    fh = self.path

            unpacker = unpack(fh)
            for o in unpacker:
                yield o
        finally:
            if needs_closing:
                fh.close()



import pandas as pd
import numpy as np
import pytest, random, string, os, datetime

from . import (
    make_string_index,
    make_integer_index,
    make_float_index,
    make_datetime_index,
    make_categorical_index,
    make_range_index,
)
    

@pytest.fixture
def pandas_data():
    d = {
        'string': make_string_index(100),
        'date': make_datetime_index(100),
        'int': make_integer_index(100),
        'rng': make_range_index(100),
        'float': make_float_index(100),
        'empty': pd.Index([]),
        'tuple': pd.Index(zip(['foo', 'bar', 'baz'], [1, 2, 3])),
        'period': pd.Index(pd.period_range('2012-1-1', freq='M', periods=3)),
        'date2': pd.Index(pd.date_range('2013-01-1', periods=10)),
        'bdate': pd.Index(pd.bdate_range('2013-01-02', periods=10)),
        'cat': make_categorical_index(100)
    }

    return d

@pytest.fixture
def pandas_multi_index():
    mi = {
        'reg': pd.MultiIndex.from_tuples(
            [('bar', 'one'), ('baz', 'two'),
            ('foo', 'two'),
            ('qux', 'one'), ('qux', 'two')],
            names=['first', 'second']),
        }
    return mi


@pytest.fixture
def pandas_categorical_data():
    d = {}
    d['plain_str'] = pd.Categorical(['a', 'b', 'c', 'd', 'e'])
    d['plain_str_ordered'] = pd.Categorical(['a', 'b', 'c', 'd', 'e'], ordered=True)
    d['plain_int'] = pd.Categorical([5, 6, 7, 8])
    d['plain_int_ordered'] = pd.Categorical([5, 6, 7, 8], ordered=True)
    return d


@pytest.fixture
def pandas_dataframe():

    # data = {
    #     'A': [0., 1., 2., 3., np.nan],
    #     'B': [0, 1, 0, 1, 0],
    #     'C': ['foo1', 'foo2', 'foo3', 'foo4', 'foo5'],
    #     'D': pd.date_range('1/1/2009', periods=5),
    #     'E': [0., 1, pd.Timestamp('20100101'), 'foo', 2.],
    #     'H': pd.Categorical(['a', 'b', 'c', 'd', 'e']),
    #     'I': pd.Categorical(['a', 'b', 'c', 'd', 'e'], ordered=True),
    # }
    data = {
        'A': [u'\u2019'] * 1000,
        'B': np.arange(1000, dtype=np.int32),
        'C': list(100 * 'abcdefghij'),
        'D': pd.date_range(datetime.datetime(2015, 4, 1), periods=1000),
        'E': [datetime.timedelta(days=x) for x in range(1000)],
        'G': [400] * 1000
    }
    frame = {
        'float': pd.DataFrame(dict(A=data['A'], B=pd.Series(data['A']))),
        'int': pd.DataFrame(dict(A=data['B'], B=pd.Series(data['B']))),
        'mixed': pd.DataFrame(data)}
    return frame

@pytest.fixture
def pandas_datetime_dataframe():
    datetime_frame = {
        "datetime": pd.DataFrame({
        'F': [pd.Timestamp('20130102', tz='US/Eastern')] * 5,
        'G': [pd.Timestamp('20130603', tz='CET')] * 5,
    })}
    return datetime_frame

def random_string(length):
   letters = string.ascii_lowercase
   return ''.join(random.choice(letters) for i in range(length))


@pytest.fixture
def tmp_file_msg(tmp_path):
    """
    Create a temporary file with the `.msg` extension.
    """
    f = '__%s__.msg' % random_string(10)
    return os.path.join(tmp_path, f)


import numpy as np
import pytest
from . import encode_decode

def test_numpy_scalar_float(tmp_file_msg):
    x = np.float32(np.random.rand())
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == pytest.approx(x_rec)

def test_numpy_scalar_complex(tmp_file_msg):
    x = np.complex64(np.random.rand() + 1j * np.random.rand())
    x_rec = encode_decode(tmp_file_msg, x)
    assert np.allclose(x, x_rec)

def test_scalar_float(tmp_file_msg):
    x = np.random.rand()
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == pytest.approx(x_rec)

def test_scalar_complex(tmp_file_msg):
    x = np.random.rand() + 1j * np.random.rand()
    x_rec = encode_decode(tmp_file_msg, x)
    assert np.allclose(x, x_rec)

def test_list_numpy_float(tmp_file_msg):
    x = [np.float32(np.random.rand()) for i in range(5)]
    x_rec = encode_decode(tmp_file_msg, x)
    # current msgpack cannot distinguish list/tuple
    assert tuple(x) == pytest.approx(x_rec)

    x_rec = encode_decode(tmp_file_msg, tuple(x))
    assert tuple(x) == pytest.approx(x_rec)

def test_list_numpy_float_complex(tmp_file_msg):
    if not hasattr(np, 'complex128'):
        pytest.skip('numpy cant handle complex128')

    x = [np.float32(np.random.rand()) for i in range(5)] + \
        [np.complex128(np.random.rand() + 1j * np.random.rand())
            for i in range(5)]
    x_rec = encode_decode(tmp_file_msg, x)
    assert np.allclose(x, x_rec)

def test_list_float(tmp_file_msg):
    x = [np.random.rand() for i in range(5)]
    x_rec = encode_decode(tmp_file_msg, x)
    # current msgpack cannot distinguish list/tuple
    assert tuple(x) == pytest.approx(x_rec)

    x_rec = encode_decode(tmp_file_msg, tuple(x))
    assert tuple(x) == pytest.approx(x_rec)

def test_list_float_complex(tmp_file_msg):
    x = [np.random.rand() for i in range(5)] + \
        [(np.random.rand() + 1j * np.random.rand()) for i in range(5)]
    x_rec = encode_decode(tmp_file_msg, x)
    assert np.allclose(x, x_rec)

def test_dict_float(tmp_file_msg):
    x = {'foo': 1.0, 'bar': 2.0}
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == pytest.approx(x_rec)

def test_dict_complex(tmp_file_msg):
    x = {'foo': 1.0 + 1.0j, 'bar': 2.0 + 2.0j}
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == x_rec
    for key in x:
        assert type(x[key]) == type(x_rec[key])

def test_dict_numpy_float(tmp_file_msg):
    x = {'foo': np.float32(1.0), 'bar': np.float32(2.0)}
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == pytest.approx(x_rec)

def test_dict_numpy_complex(tmp_file_msg):
    x = {'foo': np.complex128(1.0 + 1.0j),
            'bar': np.complex128(2.0 + 2.0j)}
    x_rec = encode_decode(tmp_file_msg, x)
    assert x == x_rec
    for key in x:
        assert type(x[key]) == type(x_rec[key])

def test_numpy_array_float(tmp_file_msg):
    # run multiple times
    for _ in range(10):
        x = np.random.rand(10)
        for dtype in ['float32', 'float64']:
            x = x.astype(dtype)
            x_rec = encode_decode(tmp_file_msg, x)
            assert x == pytest.approx(x_rec)

def test_numpy_array_complex(tmp_file_msg):
    x = (np.random.rand(5) + 1j * np.random.rand(5)).astype(np.complex128)
    x_rec = encode_decode(tmp_file_msg, x)
    assert all(map(lambda x, y: x == y, x, x_rec)) and x.dtype == x_rec.dtype

def test_list_mixed(tmp_file_msg):
    x = [1.0, np.float32(3.5), np.complex128(4.25), u'foo']
    x_rec = encode_decode(tmp_file_msg, x)
    # current msgpack cannot distinguish list/tuple
    assert tuple(x) == pytest.approx(x_rec)

    x_rec = encode_decode(tmp_file_msg, tuple(x))
    assert tuple(x) == pytest.approx(x_rec)


from . import encode_decode
from pandas._testing import assert_categorical_equal

def test_basic(pandas_categorical_data):
    # run multiple times here
    for _ in range(10):
        for s, i in pandas_categorical_data.items():
            i_rec = encode_decode(None, i)
            assert_categorical_equal(i, i_rec)


import pandas as pd
from isf_pandas_msgpack import to_msgpack, read_msgpack
from pandas._testing import (
    assert_frame_equal,
    assert_series_equal,
    assert_index_equal,
)

def encode_decode(path, x, compress=None, **kwargs):
    if path is None:
        # returns the msgpack string
        s = to_msgpack(None, x, compress=compress, **kwargs)
        return read_msgpack(s, **kwargs)
    else:
        to_msgpack(path, x, compress=compress, **kwargs)
        return read_msgpack(path, **kwargs)


def check_arbitrary(a, b):
    if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        assert(len(a) == len(b))
        for a_, b_ in zip(a, b):
            check_arbitrary(a_, b_)
    elif isinstance(a, pd.DataFrame):
        assert_frame_equal(a, b)
    elif isinstance(a, pd.Series):
        assert_series_equal(a, b)
    elif isinstance(a, pd.Index):
        assert_index_equal(a, b)
    elif a is pd.NaT:
        assert b is pd.NaT
    elif isinstance(a, pd.Timestamp):
        assert a == b
        assert a.freq == b.freq
    else:
        assert(a == b)


def make_string_index(length):
    """
    Create a string index of the given length.
    """
    return pd.Index([str(i) for i in range(length)])


def make_integer_index(length):
    """
    Create an integer index of the given length.
    """
    return pd.Index(range(length))


def make_float_index(length):
    """
    Create a float index of the given length.
    """
    return pd.Index([float(i) for i in range(length)])


def make_datetime_index(length):
    """
    Create a datetime index of the given length.
    """ 
    return pd.date_range(start='2020-01-01', periods=length, freq='D')

def make_timedelta_index(length):
    """
    Create a timedelta index of the given length.
    """
    return pd.to_timedelta(pd.date_range(start='2020-01-01',
                                          periods=length, freq='D') - pd.Timestamp('2020-01-01'))
    
    
def make_boolean_index(length):
    """
    Create a boolean index of the given length.
    """
    return pd.Index([True if i % 2 == 0 else False for i in range(length)])


def make_categorical_index(length):
    """
    Create a categorical index of the given length.
    """
    return pd.CategoricalIndex([str(i) for i in range(length)])


def make_range_index(length):
    """
    Create a range index of the given length.
    """
    return pd.RangeIndex(start=0, stop=length, step=1)

    

def make_interval_index(length):
    """
    Create an interval index of the given length.
    """
    return pd.IntervalIndex.from_breaks(range(0, length * 2, 2))


def make_period_index(length):
    """
    Create a period index of the given length.
    """
    return pd.PeriodIndex(pd.date_range(start='2020-01-01',
                                         periods=length, freq='D').to_period('D'))

def make_string_series(length=10):
    """
    Create a string series of the given length.
    """
    return pd.Series([str(i) for i in range(length)], index=make_string_index(length))


import datetime, pytest, sys
import numpy as np
import pandas as pd
from . import encode_decode

def test_timestamp():
    for i in [pd.Timestamp(
        '20130101'), pd.Timestamp('20130101', tz='US/Eastern'),
            pd.Timestamp('201301010501')]:
        i_rec = encode_decode(None, i)
        assert i == i_rec

def test_nat():
    nat_rec = encode_decode(None, pd.NaT)
    assert pd.isnull(nat_rec)

@pytest.mark.skipif(
    sys.version_info < (2, 7),
    reason='datetime64 broken in 2.6')
def test_datetimes():
    for i in [datetime.datetime(2013, 1, 1),
                datetime.datetime(2013, 1, 1, 5, 1),
                datetime.date(2013, 1, 1),
                np.datetime64(datetime.datetime(2013, 1, 5, 2, 15))]:
        i_rec = encode_decode(None, i)
        assert i == i_rec

def test_timedeltas():
    for i in [datetime.timedelta(days=1),
                datetime.timedelta(days=1, seconds=10),
                np.timedelta64(1000000)]:
        i_rec = encode_decode(None, i)
        assert i == i_rec 


from . import encode_decode, make_string_index
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal, assert_index_equal

def test_basic_index(pandas_data):

    for s, i in pandas_data.items():
        i_rec = encode_decode(None, i)
        assert_index_equal(i, i_rec)

    # datetime with no freq (GH5506)
    i = pd.Index([pd.Timestamp('20130101'), pd.Timestamp('20130103')])
    i_rec = encode_decode(None, i)
    assert_index_equal(i, i_rec)

    # datetime with timezone
    i = pd.Index([pd.Timestamp('20130101 9:00:00'), pd.Timestamp(
        '20130103 11:00:00')]).tz_localize('US/Eastern')
    i_rec = encode_decode(None, i)
    assert_index_equal(i, i_rec)

def test_multi_index(pandas_multi_index):

    for s, i in pandas_multi_index.items():
        i_rec = encode_decode(None, i)
        assert_index_equal(i, i_rec)

def test_str_index():
    i = make_string_index(100)

    i_rec = encode_decode(None, i)
    assert_index_equal(i, i_rec)

def test_categorical_index():
    # GH15487
    df = pd.DataFrame(np.random.randn(10, 2))
    df = df.astype({0: 'category'}).set_index(0)
    result = encode_decode(None, df)
    assert_frame_equal(result, df)


from pandas.testing import assert_frame_equal
from . import encode_decode, to_msgpack

UTF_ENCODINGS = ['utf8', 'utf16', 'utf32']

def test_utf(pandas_dataframe):
    # GH10581
    for encoding in UTF_ENCODINGS:
        for frame in pandas_dataframe.values():
            result = encode_decode(None, frame, encoding=encoding)
            assert_frame_equal(result, frame)

def test_default_encoding(pandas_dataframe):
    for frame in pandas_dataframe.values():
        result = to_msgpack(frame)
        expected = to_msgpack(frame, encoding='utf8')
        assert result == expected
        result = encode_decode(None, frame)
        assert_frame_equal(result, frame)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from setuptools import setup
import importlib.resources, importlib.metadata
from distutils.extension import Extension
from distutils.command.build_ext import build_ext as build_ext

NAME = 'pandas-msgpack'

def is_platform_windows():
    return sys.platform == 'win32' or sys.platform == 'cygwin'

def is_platform_linux():
    return sys.platform == 'linux2'

def is_platform_mac():
    return sys.platform == 'darwin'

try:
    import Cython
    from Cython.Build import cythonize
except ImportError:
    raise ImportError("cython is required for building")

# args to ignore warnings
if is_platform_windows():
    extra_compile_args=[]
else:
    extra_compile_args=['-Wno-unused-function']


if sys.byteorder == 'big':
    macros = [('__BIG_ENDIAN__', '1')]
else:
    macros = [('__LITTLE_ENDIAN__', '1')]

extensions = []
packer_ext = Extension('isf_pandas_msgpack.msgpack._packer',
                        depends=['isf_pandas_msgpack/includes/pack.h',
                                 'isf_pandas_msgpack/includes/pack_template.h'],
                        sources = ['isf_pandas_msgpack/msgpack/_packer.pyx'],
                        language='c++',
                        include_dirs=['isf_pandas_msgack/includes'],
                        define_macros=macros,
                        extra_compile_args=extra_compile_args)
unpacker_ext = Extension('isf_pandas_msgpack.msgpack._unpacker',
                        depends=['isf_pandas_msgpack/includes/unpack.h',
                                 'isf_pandas_msgpack/includes/unpack_define.h',
                                 'isf_pandas_msgpack/includes/unpack_template.h'],
                        sources = ['isf_pandas_msgpack/msgpack/_unpacker.pyx'],
                         language='c++',
                        include_dirs=['isf_pandas_msgpack/includes'],
                        define_macros=macros,
                        extra_compile_args=extra_compile_args)
extensions.append(packer_ext)
extensions.append(unpacker_ext)

#----------------------------------------------------------------------
# util
# extension for pseudo-safely moving bytes into mutable buffers
_move_ext = Extension('isf_pandas_msgpack._move',
                      depends=[],
                      sources=['isf_pandas_msgpack/move.c'])
extensions.append(_move_ext)

setup(
    name=NAME,
    description="Pandas interface to msgpack",
    ext_modules=cythonize(extensions),
    packages=['isf_pandas_msgpack',
              'isf_pandas_msgpack.includes',
              'isf_pandas_msgpack.msgpack',
              'isf_pandas_msgpack.tests'],
)
